/* ------------------------------------------------------------------------- *\
 * BinarySearchTree
 * Interface of a binary search tree particularized for storing "location" 
 *(in a broad sense) of strings.
 *
 * Key : string (char*)
 * Values : x (see X.h) for storing locations (size_t)
 *
 * Note: You are free to define BinarySearchTree_t in any way you want. 
 * Note: You can expect that locations will be given in ascending order with
 *  respect to a given string. A valid insertion sequence is thus :
 *      ("A", 1)
 *      ("B", 7)
 *      ("A", 2)

\* ------------------------------------------------------------------------- */

#ifndef _BINARY_SEARCH_TREE_H_
#define _BINARY_SEARCH_TREE_H_

#include <stddef.h>
#include <stdbool.h>

typedef enum {NEW_TREE_KEY, ALREADY_PRESENT_IN_TREE, TREE_ALLOC_ERROR} insertTree_t;


// Opaque structure of a binary search tree.
typedef struct BinarySearchTree_t BinarySearchTree;

/* ------------------------------------------------------------------------- *\
 * Create an empty binary search tree.
 * The binary search tree must later be deleted by calling
 * freeBinarySearchTree().
 *
 * RETURN
 * tree         Pointer to an empty binary search tree
 * NULL         if an error occured
\* ------------------------------------------------------------------------- */
BinarySearchTree* createBinarySearchTree(void);

/* ------------------------------------------------------------------------- *\
 * Destroy a binary search tree that has been created by
 * createBinarySearchTree(). 
 *
 * PARAMETER
 * tree         Pointer to a binary search tree
 * freeContent  Whether to free directly the content or not. If true,
 *              the string serving as key must be freed (the data structure
 *              holding the locations must always be cleared)
\* ------------------------------------------------------------------------- */
void freeBinarySearchTree(BinarySearchTree* tree, bool freeContent);

/* ------------------------------------------------------------------------- *\
 * Return the number of different strings in a binary search tree (i.e. the
 * number of keys).
 *
 * PARAMETER
 * tree         Pointer to a binary search tree
 *
 * RETURN
 * size         Number of different strings stored in the tree
\* ------------------------------------------------------------------------- */
size_t getNumString(const BinarySearchTree* tree);


/* ------------------------------------------------------------------------- *\
 * Return the total number of strings in a binary search tree (i.e. the sum
 * of each location for every string).
 *
 * PARAMETER
 * tree         Pointer to a binary search tree
 *
 * RETURN
 * size         Total number of strings stored in the tree
\* ------------------------------------------------------------------------- */
size_t getTotalNumString(const BinarySearchTree* tree);

/* ------------------------------------------------------------------------- *\
 * Insert a new location for the given string. Location must be given in
 * ascending order with respect to a same string.
 *
 * PARAMETERS
 * tree         Pointer to a binary search tree
 * string       string whose location to store
 * location     location of the string to store
 *
 * RETURN
 * inserted     Whether the given string is a new key or not, or if an error
 *				occured
\* ------------------------------------------------------------------------- */
insertTree_t insert(BinarySearchTree* tree, char* string, size_t location);


/* ------------------------------------------------------------------------- *\
 * Find all the locations of the string in a binary search tree.
 *
 * PARAMETER
 * tree                 Pointer to a binary search tree
 * string               The string to search
 * locationsPointer     A valid pointer to a pointer of an array of size_t
 *                      The pointer is set to NULL in case of allocation error
 *                      or
 *                      if the search was unsuccessful
 *
 * Example of usage:
 * char* string = "geometry";
 * size_t* locations;
 * size_t nbLocation = find(myTree, string, &locations);
 * printf("%zu", locations[0]); // Prints the first location (you must 
 *                              // check for error)
 *
 * NOTE
 *  - It is the responsibility of the caller to clear the array pointed
 *    to by locations
 *
 * RETURN
 * nbLocations  The number of locations in which the string appears
 *              or
 *              - 1 is case of error
\* ------------------------------------------------------------------------- */
int find(const BinarySearchTree* tree, const char* string, size_t** locationsPointer);

/* ------------------------------------------------------------------------- *\
 * Find all the locations where all the strings appear together.
 *
 * PARAMETER
 * tree                 Pointer to a binary search tree
 * strings              An array of strings
 * nbStrings            The number of strings in the array
 * locationsPointer     A valid pointer to a pointer of a array of size_t
 *                      The pointer is set to NULL in case of allocation error
 *                      or
 *                      if the search was unsuccessful
 *
 * NOTE
 *  - It is the responsibility of the caller to clear the array pointed
 *    to by locations
 *
 * RETURN
 * nbLocations  The number of locations in which the string co-appears
 *              or
 *              -1 in case of error
\* ------------------------------------------------------------------------- */
int findCooccurrences(const BinarySearchTree* tree, char** strings, size_t nbStrings, size_t** locationsPointer);

/* ------------------------------------------------------------------------- *\
 * Check whether a binary search tree is balanced.

 * A tree is considered balanced if for all nodes in the tree
 * - both the left and right sub-trees are balanced;
 * - the difference between the heights of the left and right sub-trees
 *   is at most 1.
 *
 * PARAMETERS
 * tree         Pointer to a binary search tree
 *
 * RETURN
 * true         If the binary search is balanced
 * false        Otherwise
\* ------------------------------------------------------------------------- */
bool isBalanced(const BinarySearchTree* tree);


/* ------------------------------------------------------------------------- *\
 * Return the height of the tree.
 *
 * PARAMETERS
 * tree         Pointer to a binary search tree
 *
 * RETURN
 * height       The height of the tree
\* ------------------------------------------------------------------------- */
int getHeight(const BinarySearchTree* tree);


/* ------------------------------------------------------------------------- *\
 * Return the average depth of the tree
 *
 * PARAMETERS
 * tree         Pointer to a binary search tree
 *
 * RETURN
 * avgDepth     The average depth of the tree
\* ------------------------------------------------------------------------- */
double getAvgDepth(const BinarySearchTree* tree);



#endif // _BINARY_SEARCH_TREE_H_

