
/* ------------------------------------------------------------------------- *\
 * SequentialSet :
 * Interface of a sequential set. I.e. a set (no duplicates) which expect
 * value (size_t) insertions in increasing order.
\* ------------------------------------------------------------------------- */
#ifndef _SEQUENTIAL_SET_H_
#define _SEQUENTIAL_SET_H_

#include <stdbool.h>
 
typedef enum {INSERTED, ALREADY_PRESENT, ALLOC_ERROR} insert_t;

// Opaque structure for sequential sets
typedef struct SequentialSet_t SequentialSet;

/* ------------------------------------------------------------------------- *\
 * Create an empty sequential set.
 * The sequential set must later be deleted by calling
 * freeSequentialSet().
 *
 * RETURN
 * seqSet       Pointer to an empty sequential set
 * NULL         if an error occured
\* ------------------------------------------------------------------------- */
SequentialSet* createSequentialSet(void);

/* ------------------------------------------------------------------------- *\
 * Destroy a sequential set that has been created by createSequentialSet(). 
 * 
 * PARAMETER
 * seqSet       Pointer to a sequential set
\* ------------------------------------------------------------------------- */
void freeSequentialSet(SequentialSet* seqSet);

/* ------------------------------------------------------------------------- *\
 * Return the number of items in the sequential set.
 *
 * PARAMETER
 * seqSet       Pointer to a sequential set
 *
 * RETURN
 * size         Number of items
\* ------------------------------------------------------------------------- */
size_t getNumItems(const SequentialSet* seqSet);


/* ------------------------------------------------------------------------- *\
 * Insert a new value in the sequential set.
 *
 * PARAMETERS
 * seqSet       Pointer to a sequential set
 * value        value to store
 *
 * RETURN
 * status		An indication of the insertion result   
 * 				- If the value is already present, return ALREADY_PRESENT
 *  			- If the value is inserted, return INSERTED
 *				- If an allocation error occured, return ALLOC_ERROR
\* ------------------------------------------------------------------------- */
insert_t insertValue(SequentialSet* seqSet, size_t value);



/* ------------------------------------------------------------------------- *\
 * Checks whether the values is already present in the set
 *
 * PARAMETERS
 * seqSet       Pointer to a sequential set
 * value        value to check
 *
 * RETURN
 * contains     Whether the given value is in the set
\* ------------------------------------------------------------------------- */
bool contains(SequentialSet* seqSet, size_t value);


/* ------------------------------------------------------------------------- *\
 * Return an instantaneous array view of the set (in ascending order).
 * Use getNumItems(seqSet) to get the length of the array.
 * 
 * NOTE
 *  - It is the responsibility of the caller to clear the array
 *
 * PARAMETER
 * seqSet       Pointer to a sequential set
 *
 * RETURN
 * setArray     The array corresponding to the set
 *              NULL if an error occured
\* ------------------------------------------------------------------------- */
size_t* setAsArray(const SequentialSet* seqSet);


/* ------------------------------------------------------------------------- *\
 * Return an new sequential set corresponding to the intersection (all and
 * only all the values present in all the sets) of the given sequential sets.
 * 
 * NOTE
 *  - It is the responsibility of the caller to clear the returned sequential
 *    set
 *
 * PARAMETER
 * seqSetArray  Pointer to an array of sequential sets
 * nbSets       The number of sets
 *
 * RETURN
 * seqSet       The sequential set corresponding to the intersection
\* ------------------------------------------------------------------------- */
SequentialSet* intersect(SequentialSet** seqSetArray, size_t nbSets);





#endif // _SEQUENTIAL_SET_H_
