#include <stddef.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "BinarySearchTree.h"

// File reading buffer maximum capacity
static const size_t BUFFER_SIZE = 4096;

// Maximum number of line to print
static const size_t MAX_PRINT_LINE = 10;

/* ------------------------------------------------------------------------- *\
 * Extract the substring ranging from the start of the string up to the 
 * delimiter (excluded).
 * 
 * Remarks:
 * - The original string is modified. Consider giving a copy and saving
 *   a pointer to the start of the string (memory leaks avoidance)
 * - The substring returned is allocated dynamially. Do not forget
 *   to free it !
 *
 * PARAMETERS
 * stringPointer    Pointer to the starting pointer of the string
 * delim            The character which delimits parts of the string
 *
 * RETURN
 * substr           Pointer to the substring
 * NULL             if stringPointer is null, an error occured or the string
 *                  is empty (only the null character)
\* ------------------------------------------------------------------------- */
static char* _strsep(char** stringPointer, char delim)
{

    if (*stringPointer == NULL)
        return NULL;

    char* string = *stringPointer;
    char* substr = NULL;

    // Test whether the string is empty
    if (*string == 0)
        return NULL;

    //Removing trailing delimiters
    while(*string == delim)
        string++;
    
    // Process the string in search of the delimiter
    size_t i = 0;
    while (string[++i] != 0)
    {
        if(string[i] == delim)
            break;   
    }

    // Allocate the required memory, copy the appropriate part and return
    *stringPointer = string + i;
    substr = calloc((i+1), sizeof(char));
    if (substr == NULL)
        return NULL;
    strncpy(substr, string, i);
    return substr;

}


/* ------------------------------------------------------------------------- *\ 
 * NOM
 *      find_occurrences
 * SYNOPSIS
 *      find_occurrences [-b] file [word1 [word2 ...]]
 * DESCIRPTION
 *      find lines of (co)occurrences of several words.
 * OPTION
 *      -b      Whether to keep the underlying tree balanced
 * USAGE
 *      ./find_occurrences time_machine_formated.txt the a
 *          will find the line where both "the" and "a" appear.
 \* ------------------------------------------------------------------------- */
int main(int argc, char** argv)
{
    // Default arguments
    size_t nbWordsToFind = 0;
    char** wordsToFind = NULL;
    size_t argIndex = 1;
    char* file = NULL;

    /* ---------- Command line part -----------*/
    // Parsing command line arguments
    if (argc < 2)
    {
        fprintf(stderr, "Program is expecting at least 1 arguments : the path to the file to process\n");
        exit(EXIT_FAILURE);
    }
    file = argv[argIndex++];
    if ((int)argIndex < argc)//More arguments available
    {
        nbWordsToFind = argc - argIndex;
        wordsToFind = argv + argIndex;
    }


    /* ---------- Resource creation -----------*/

    // Open the file
    FILE* fd = fopen(file, "r");
    if (fd == NULL)
    {
        fprintf(stderr, "Could not open file '%s'. Exiting...\n", file);
        exit(EXIT_FAILURE);
    }

    // Allocate memory for dynamic buffer
    char* sentence = NULL;
    sentence = malloc((BUFFER_SIZE + 1) * sizeof(char));// Dynamic buffer
    if (sentence == NULL)
    {
        fprintf(stderr, "Memory error. Exiting...\n");
        fclose(fd);
        exit(EXIT_FAILURE);
    }

    // Create an empty tree
    BinarySearchTree* bst = createBinarySearchTree();// The dictionary
    if (bst == NULL)
    {
        fprintf(stderr, "Memory error. Exiting...\n");
        fclose(fd);
        free(sentence);
        exit(EXIT_FAILURE);
    }

    char* startSentence = sentence;// Pointer to the start of the dynamic buffer
    char* word = NULL;// A word to add

    char buffer[BUFFER_SIZE];// Static buffer for reading the file

    size_t sentenceNb = 0;// The number of sentences


    /* ---------- File scanning and tree filling -----------*/

    // Go through the file and extract the words of each sentence/line (while monitoring time)
    printf("Building the tree...\n");
    clock_t startBuilding = clock();
    while (fgets(buffer, sizeof(buffer), fd) != NULL) 
    {
        sentence = startSentence; //Reset the pointer at the start of the buffer
        sscanf(buffer, "%[^\n]", sentence);
        if (sentence != NULL) 
        {
            // Decompose the sentence into its words
            while ((word = _strsep(&sentence, ' ')) != NULL)
            {                
                if(insert(bst, word, sentenceNb) != NEW_TREE_KEY)
                    free(word);//If the word was not inserted it must be freed now or the pointer will be lost
            }
        }
        sentenceNb++;
    }

    double buildingTime = ((double) (clock() - startBuilding)) / CLOCKS_PER_SEC;

    // Free resources (release file handle and dynamic buffer)
    fclose(fd);
    free(startSentence);


    /* ---------- Dictionary application -----------*/

    printf("==== Stats ====\n");
    printf("Time to build the tree : %f sec.\n", buildingTime);
    printf("Number of different words : %zu\n", getNumString(bst));
    printf("Total number of words : %zu\n", getTotalNumString(bst));
    printf("Tree height : %zu\n", getHeight(bst));
    printf("Tree average depth : %f\n", getAvgDepth(bst));
    printf("The tree is %s\n", isBalanced(bst) ? "balanced" : "unbalanced");

    if (nbWordsToFind > 0)
    {
        printf("==== Search ====\n");
        printf("Looking for words : ");
        for (size_t i=0; i<nbWordsToFind; i++)
        {
            printf("'%s' ", wordsToFind[i]);
        }
        printf("\n");

        size_t* sentences = NULL;
        int nbSentences = 0;
        clock_t startSearching = clock();
        if (nbWordsToFind == 1)
        {

            nbSentences = find(bst, *wordsToFind, &sentences);
            
        } else { // Several words
        
            nbSentences = findCooccurrences(bst, wordsToFind, nbWordsToFind, &sentences);

        }
        double searchTime = ((double) (clock() - startSearching)) / CLOCKS_PER_SEC;
        switch (nbSentences)
        {
            case -1:
                fprintf(stderr, " -- An error occured in the search phase\n");
                break;
            case 0:
                printf(" -- Word(s) not found\n");
                break;
            default:
                // Print the first MAX_PRINT_LINE of the (co)ocurrence
                for (size_t i=0; i<(size_t)nbSentences && i<MAX_PRINT_LINE; i++)
                    printf(" -- Word(s) found in sentence(s) %zu\n", sentences[i]);
                if ((size_t)nbSentences > MAX_PRINT_LINE) 
                    printf(" -- ...\n");
                printf("%d (co)ocurrence(s) found\n", nbSentences);
                break;
        }
        free(sentences); // Free resources
        printf("Search time : %f sec.\n", searchTime);
    }


    /* ---------- Final releases -----------*/

    // Free resources 
    freeBinarySearchTree(bst, true); //free BST and all its content
    
    exit(EXIT_SUCCESS);
}
